@@ -7,6 +7,7 @@
 #  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
 import dataclasses
 import types
+import typing
 from typing import Any, Type, Callable
 
 from init_helpers import is_instance
@@ -16,20 +17,22 @@ from more_itertools import first
 from .defs import Json
 from .typed_In_dumps import TypedInDumps
 
-LiteralTypeWrapper = type[TypedInDumps] | Callable[[dict | list | str | float | int | bool | None], Any]
+UntypedValueWrapper = type[TypedInDumps] | Callable[[dict | list | str | float | int | bool | None], Any]
 
 
 class TypedJsonParser:
-    def __init__(self, literal_type_wrapper: LiteralTypeWrapper = lambda x: x, skip_unknown_t: bool = False):
-        self._literal_type_wrapper = literal_type_wrapper
+    def __init__(self, untyped_value_wrapper: UntypedValueWrapper = lambda x: x, skip_unknown_t: bool = False):
+        self._untyped_value_wrapper = untyped_value_wrapper
         self._wrap_literal = self._prepare_wrap_literal()
         self.skip_unknown_t = skip_unknown_t
 
     def _prepare_wrap_literal(self) -> Callable[[dict | list | str | float | int | bool | None], Any]:
-        if not isinstance(self._literal_type_wrapper, type) or not issubclass(self._literal_type_wrapper, TypedInDumps):
-            return self._literal_type_wrapper
-
-        wrapper_field_name_to_field = get_dataclass_field_name_to_field(self._literal_type_wrapper)
+        if not isinstance(self._untyped_value_wrapper, type):
+            return self._untyped_value_wrapper
+        if not issubclass(self._untyped_value_wrapper, TypedInDumps):
+            return self._untyped_value_wrapper
+        # todo: assert is_dataclass_type(self._untyped_value_wrapper)
+        wrapper_field_name_to_field = get_dataclass_field_name_to_field(self._untyped_value_wrapper)
         wrapper_field_name_to_field = {name: field for name, field in wrapper_field_name_to_field.items() if field.init}
         if len(wrapper_field_name_to_field) != 1:
             wrapper_field_name_to_field = {
@@ -43,7 +46,7 @@ class TypedJsonParser:
                 "it should have one field with init=True and no defaults")
 
         field_name = first(wrapper_field_name_to_field)
-        return lambda x: self._literal_type_wrapper(**{field_name: x})
+        return lambda x: self._untyped_value_wrapper(**{field_name: x})
 
     def _wrap_literal(
             self,
@@ -52,10 +55,11 @@ class TypedJsonParser:
         pass
 
     def parse(self, value: Any, target_type: Type[TypedInDumps | Json] | types.UnionType = TypedInDumps):
-        if is_instance(value, [target_type]):
-            return value
+        if isinstance(target_type, (type, types.UnionType)) or getattr(target_type, '__origin__', None) is typing.Union:
+            if is_instance(value, [target_type]):
+                return value
 
-        if target_type is self._literal_type_wrapper:
+        if target_type is self._untyped_value_wrapper:
             if not isinstance(value, TypedInDumps):
                 value = self.parse(
                     value, dict[str, TypedInDumps] | list[TypedInDumps] | str | float | int | bool | None)
@@ -64,9 +68,9 @@ class TypedJsonParser:
         if isinstance(target_type, type) and issubclass(target_type, TypedInDumps):
             if isinstance(value, target_type):
                 return value
+            prefixed_name_to_non_abstract_type = target_type.get_prefixed_name_to_subclass()
             if isinstance(value, dict) and '_t' in value:
                 type_name = value.get('_t')
-                prefixed_name_to_non_abstract_type = target_type.get_prefixed_name_to_non_abstract_subclass()
                 try:
                     type_ = prefixed_name_to_non_abstract_type[type_name]  # TODO add some kind of skip for unknown
                 except (KeyError, TypeError) as e:
@@ -86,7 +90,14 @@ class TypedJsonParser:
                         if val is not NoValue:
                             kwargs[key] = val
                     return type_(**kwargs)
-            return self.parse(value, self._literal_type_wrapper)
+
+            result = self.parse(value, self._untyped_value_wrapper)
+            if not issubclass(type(result), target_type):
+                raise ValueError(
+                    f"Required {target_type.__name__}({list(prefixed_name_to_non_abstract_type)}), "
+                    f"got untyped {type(value).__name__}, applied untyped_wrapper, got: {type(result).__name__}"
+                )
+            return result
 
         target_type_origin = getattr(target_type, '__origin__', None)
         target_type_args = getattr(target_type, '__args__', None)
